#                      Data Parallel Control (dpctl)
#
# Copyright 2020-2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import dpctl


def has_gpu(backend="opencl"):
    return bool(dpctl.get_num_devices(backend=backend, device_type="gpu"))


def has_cpu(backend="opencl"):
    return bool(dpctl.get_num_devices(backend=backend, device_type="cpu"))


def has_sycl_platforms():
    return bool(len(dpctl.get_platforms()))


def create_invalid_capsule():
    """Creates an invalid capsule for the purpose of testing dpctl
    constructors that accept capsules.
    """
    import ctypes

    ctor = ctypes.pythonapi.PyCapsule_New
    ctor.restype = ctypes.py_object
    ctor.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_void_p]
    return ctor(id(ctor), b"invalid", 0)


def get_queue_or_skip(args=tuple()):
    try:
        q = dpctl.SyclQueue(*args)
    except dpctl.SyclQueueCreationError:
        pytest.skip(f"Queue could not be created from {args}")
    return q


def skip_if_dtype_not_supported(dt, q_or_dev):
    import dpctl.tensor as dpt

    dt = dpt.dtype(dt)
    if type(q_or_dev) is dpctl.SyclQueue:
        dev = q_or_dev.sycl_device
    elif type(q_or_dev) is dpctl.SyclDevice:
        dev = q_or_dev
    else:
        raise TypeError(
            "Expected dpctl.SyclQueue or dpctl.SyclDevice, "
            f"got {type(q_or_dev)}"
        )
    dev_has_dp = dev.has_aspect_fp64
    if dev_has_dp is False and dt in [dpt.float64, dpt.complex128]:
        pytest.skip(
            f"{dev.name} does not support double precision floating point types"
        )
    dev_has_hp = dev.has_aspect_fp16
    if dev_has_hp is False and dt in [
        dpt.float16,
    ]:
        pytest.skip(
            f"{dev.name} does not support half precision floating point type"
        )



#                      Data Parallel Control (dpctl)
#
# Copyright 2020-2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines unit test cases for the SyclProgram and SyclKernel classes
"""

import os

import pytest

import dpctl
import dpctl.program as dpctl_prog


def get_spirv_abspath(fn):
    curr_dir = os.path.dirname(os.path.abspath(__file__))
    spirv_file = os.path.join(curr_dir, "input_files", fn)
    return spirv_file


def _check_cpython_api_SyclProgram_GetKernelBundleRef(sycl_prog):
    """Checks Cython-generated C-API function
    `SyclProgram_GetKernelBundleRef` defined in _program.pyx"""
    import ctypes
    import sys

    assert type(sycl_prog) is dpctl_prog.SyclProgram
    mod = sys.modules[sycl_prog.__class__.__module__]
    # get capsule storing SyclProgram_GetKernelBundleRef function ptr
    kb_ref_fn_cap = mod.__pyx_capi__["SyclProgram_GetKernelBundleRef"]
    # construct Python callable to invoke "SyclProgram_GetKernelBundleRef"
    cap_ptr_fn = ctypes.pythonapi.PyCapsule_GetPointer
    cap_ptr_fn.restype = ctypes.c_void_p
    cap_ptr_fn.argtypes = [ctypes.py_object, ctypes.c_char_p]
    kb_ref_fn_ptr = cap_ptr_fn(
        kb_ref_fn_cap,
        b"DPCTLSyclKernelBundleRef (struct PySyclProgramObject *)",
    )
    # PYFUNCTYPE(result_type, *arg_types)
    callable_maker = ctypes.PYFUNCTYPE(ctypes.c_void_p, ctypes.py_object)
    get_kernel_bundle_ref_fn = callable_maker(kb_ref_fn_ptr)

    r2 = sycl_prog.addressof_ref()
    r1 = get_kernel_bundle_ref_fn(sycl_prog)
    assert r1 == r2


def _check_cpython_api_SyclProgram_Make(sycl_prog):
    """Checks Cython-generated C-API function
    `SyclProgram_Make` defined in _program.pyx"""
    import ctypes
    import sys

    assert type(sycl_prog) is dpctl_prog.SyclProgram
    mod = sys.modules[sycl_prog.__class__.__module__]
    # get capsule storing SyclProgram_Make function ptr
    make_prog_fn_cap = mod.__pyx_capi__["SyclProgram_Make"]
    # construct Python callable to invoke "SyclProgram_Make"
    cap_ptr_fn = ctypes.pythonapi.PyCapsule_GetPointer
    cap_ptr_fn.restype = ctypes.c_void_p
    cap_ptr_fn.argtypes = [ctypes.py_object, ctypes.c_char_p]
    make_prog_fn_ptr = cap_ptr_fn(
        make_prog_fn_cap,
        b"struct PySyclProgramObject *(DPCTLSyclKernelBundleRef)",
    )
    # PYFUNCTYPE(result_type, *arg_types)
    callable_maker = ctypes.PYFUNCTYPE(ctypes.py_object, ctypes.c_void_p)
    make_prog_fn = callable_maker(make_prog_fn_ptr)

    p2 = make_prog_fn(sycl_prog.addressof_ref())
    assert p2.has_sycl_kernel("add")
    assert p2.has_sycl_kernel("axpy")


def _check_cpython_api_SyclKernel_GetKernelRef(krn):
    """Checks Cython-generated C-API function
    `SyclKernel_GetKernelRef` defined in _program.pyx"""
    import ctypes
    import sys

    assert type(krn) is dpctl_prog.SyclKernel
    mod = sys.modules[krn.__class__.__module__]
    # get capsule storing SyclKernel_GetKernelRef function ptr
    k_ref_fn_cap = mod.__pyx_capi__["SyclKernel_GetKernelRef"]
    # construct Python callable to invoke "SyclKernel_GetKernelRef"
    cap_ptr_fn = ctypes.pythonapi.PyCapsule_GetPointer
    cap_ptr_fn.restype = ctypes.c_void_p
    cap_ptr_fn.argtypes = [ctypes.py_object, ctypes.c_char_p]
    k_ref_fn_ptr = cap_ptr_fn(
        k_ref_fn_cap, b"DPCTLSyclKernelRef (struct PySyclKernelObject *)"
    )
    # PYFUNCTYPE(result_type, *arg_types)
    callable_maker = ctypes.PYFUNCTYPE(ctypes.c_void_p, ctypes.py_object)
    get_kernel_ref_fn = callable_maker(k_ref_fn_ptr)

    r2 = krn.addressof_ref()
    r1 = get_kernel_ref_fn(krn)
    assert r1 == r2


def _check_cpython_api_SyclKernel_Make(krn):
    """Checks Cython-generated C-API function
    `SyclKernel_Make` defined in _program.pyx"""
    import ctypes
    import sys

    assert type(krn) is dpctl_prog.SyclKernel
    mod = sys.modules[krn.__class__.__module__]
    # get capsule storing SyclKernel_Make function ptr
    k_make_fn_cap = mod.__pyx_capi__["SyclKernel_Make"]
    # construct Python callable to invoke "SyclKernel_Make"
    cap_ptr_fn = ctypes.pythonapi.PyCapsule_GetPointer
    cap_ptr_fn.restype = ctypes.c_void_p
    cap_ptr_fn.argtypes = [ctypes.py_object, ctypes.c_char_p]
    k_make_fn_ptr = cap_ptr_fn(
        k_make_fn_cap,
        b"struct PySyclKernelObject *(DPCTLSyclKernelRef, char const *)",
    )
    # PYFUNCTYPE(result_type, *arg_types)
    callable_maker = ctypes.PYFUNCTYPE(
        ctypes.py_object, ctypes.c_void_p, ctypes.c_void_p
    )
    make_kernel_fn = callable_maker(k_make_fn_ptr)

    k2 = make_kernel_fn(
        krn.addressof_ref(), bytes(krn.get_function_name(), "utf-8")
    )
    assert krn.get_function_name() == k2.get_function_name()
    assert krn.get_num_args() == k2.get_num_args()
    assert krn.work_group_size == k2.work_group_size

    k3 = make_kernel_fn(krn.addressof_ref(), ctypes.c_void_p(None))
    assert k3.get_function_name() == "default_name"
    assert krn.get_num_args() == k3.get_num_args()
    assert krn.work_group_size == k3.work_group_size


def _check_multi_kernel_program(prog):
    assert type(prog) is dpctl_prog.SyclProgram

    assert type(prog.addressof_ref()) is int
    assert prog.has_sycl_kernel("add")
    assert prog.has_sycl_kernel("axpy")

    addKernel = prog.get_sycl_kernel("add")
    axpyKernel = prog.get_sycl_kernel("axpy")

    assert "add" == addKernel.get_function_name()
    assert "axpy" == axpyKernel.get_function_name()
    assert 3 == addKernel.get_num_args()
    assert 4 == axpyKernel.get_num_args()
    assert type(addKernel.addressof_ref()) is int
    assert type(axpyKernel.addressof_ref()) is int

    for krn in [addKernel, axpyKernel]:
        _check_cpython_api_SyclKernel_GetKernelRef(krn)
        _check_cpython_api_SyclKernel_Make(krn)

        na = krn.num_args
        assert na == krn.get_num_args()
        wgsz = krn.work_group_size
        assert type(wgsz) is int
        pwgszm = krn.preferred_work_group_size_multiple
        assert type(pwgszm) is int
        pmsz = krn.private_mem_size
        assert type(pmsz) is int
        vmnsg = krn.max_num_sub_groups
        assert type(vmnsg) is int
        v = krn.max_sub_group_size
        assert type(v) is int
        cmnsg = krn.compile_num_sub_groups
        assert type(cmnsg) is int
        cmsgsz = krn.compile_sub_group_size
        assert type(cmsgsz) is int

    _check_cpython_api_SyclProgram_GetKernelBundleRef(prog)
    _check_cpython_api_SyclProgram_Make(prog)


def test_create_program_from_source_ocl():
    oclSrc = "                                                             \
    kernel void add(global int* a, global int* b, global int* c) {         \
        size_t index = get_global_id(0);                                   \
        c[index] = a[index] + b[index];                                    \
    }                                                                      \
    kernel void axpy(global int* a, global int* b, global int* c, int d) { \
        size_t index = get_global_id(0);                                   \
        c[index] = a[index] + d*b[index];                                  \
    }"
    try:
        q = dpctl.SyclQueue("opencl")
    except dpctl.SyclQueueCreationError:
        pytest.skip("No OpenCL queue is available")
    prog = dpctl_prog.create_program_from_source(q, oclSrc)
    _check_multi_kernel_program(prog)


def test_create_program_from_spirv_ocl():
    try:
        q = dpctl.SyclQueue("opencl")
    except dpctl.SyclQueueCreationError:
        pytest.skip("No OpenCL queue is available")
    spirv_file = get_spirv_abspath("multi_kernel.spv")
    with open(spirv_file, "rb") as fin:
        spirv = fin.read()
    prog = dpctl_prog.create_program_from_spirv(q, spirv)
    _check_multi_kernel_program(prog)


def test_create_program_from_spirv_l0():
    try:
        q = dpctl.SyclQueue("level_zero")
    except dpctl.SyclQueueCreationError:
        pytest.skip("No Level-zero queue is available")
    spirv_file = get_spirv_abspath("multi_kernel.spv")
    with open(spirv_file, "rb") as fin:
        spirv = fin.read()
    prog = dpctl_prog.create_program_from_spirv(q, spirv)
    _check_multi_kernel_program(prog)


@pytest.mark.xfail(
    reason="Level-zero backend does not support compilation from source"
)
def test_create_program_from_source_l0():
    try:
        q = dpctl.SyclQueue("level_zero")
    except dpctl.SyclQueueCreationError:
        pytest.skip("No Level-zero queue is available")
    oclSrc = "                                                             \
    kernel void add(global int* a, global int* b, global int* c) {         \
        size_t index = get_global_id(0);                                   \
        c[index] = a[index] + b[index];                                    \
    }                                                                      \
    kernel void axpy(global int* a, global int* b, global int* c, int d) { \
        size_t index = get_global_id(0);                                   \
        c[index] = a[index] + d*b[index];                                  \
    }"
    prog = dpctl_prog.create_program_from_source(q, oclSrc)
    _check_multi_kernel_program(prog)


def test_create_program_from_invalid_src_ocl():
    try:
        q = dpctl.SyclQueue("opencl")
    except dpctl.SyclQueueCreationError:
        pytest.skip("No OpenCL queue is available")
    invalid_oclSrc = "                                                     \
    kernel void add(                                                       \
    }"
    with pytest.raises(dpctl_prog.SyclProgramCompilationError):
        dpctl_prog.create_program_from_source(q, invalid_oclSrc)



#                       Data Parallel Control (dpctl)
#
#  Copyright 2020-2022 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import contextlib
import operator

import numpy as np

import dpctl.tensor as dpt

__doc__ = "Print functions for :class:`dpctl.tensor.usm_ndarray`."

_print_options = {
    "linewidth": 75,
    "edgeitems": 3,
    "threshold": 1000,
    "precision": 8,
    "floatmode": "maxprec",
    "suppress": False,
    "nanstr": "nan",
    "infstr": "inf",
    "sign": "-",
}


def _options_dict(
    linewidth=None,
    edgeitems=None,
    threshold=None,
    precision=None,
    floatmode=None,
    suppress=None,
    nanstr=None,
    infstr=None,
    sign=None,
    numpy=False,
):
    if numpy:
        numpy_options = np.get_printoptions()
        options = {k: numpy_options[k] for k in _print_options.keys()}
    else:
        options = _print_options.copy()

    if suppress:
        options["suppress"] = True

    local = dict(locals().items())
    for int_arg in ["linewidth", "precision", "threshold", "edgeitems"]:
        val = local[int_arg]
        if val is not None:
            options[int_arg] = operator.index(val)

    for str_arg in ["nanstr", "infstr"]:
        val = local[str_arg]
        if val is not None:
            if not isinstance(val, str):
                raise TypeError(
                    "`{}` ".format(str_arg) + "must be of `string` type."
                )
            options[str_arg] = val

    signs = ["-", "+", " "]
    if sign is not None:
        if sign not in signs:
            raise ValueError(
                "`sign` must be one of"
                + ", ".join("`{}`".format(s) for s in signs)
            )
        options["sign"] = sign

    floatmodes = ["fixed", "unique", "maxprec", "maxprec_equal"]
    if floatmode is not None:
        if floatmode not in floatmodes:
            raise ValueError(
                "`floatmode` must be one of"
                + ", ".join("`{}`".format(m) for m in floatmodes)
            )
        options["floatmode"] = floatmode

    return options


def set_print_options(
    linewidth=None,
    edgeitems=None,
    threshold=None,
    precision=None,
    floatmode=None,
    suppress=None,
    nanstr=None,
    infstr=None,
    sign=None,
    numpy=False,
):
    """
    set_print_options(linewidth=None, edgeitems=None, threshold=None,
                      precision=None, floatmode=None, suppress=None,
                      nanstr=None, infstr=None, sign=None, numpy=False)

    Set options for printing ``dpctl.tensor.usm_ndarray`` class.

    Args:
        linewidth (int, optional): Number of characters printed per line.
            Raises `TypeError` if linewidth is not an integer.
            Default: `75`.
        edgeitems (int, optional): Number of elements at the beginning and end
            when the printed array is abbreviated.
            Raises `TypeError` if edgeitems is not an integer.
            Default: `3`.
        threshold (int, optional): Number of elements that triggers array
            abbreviation.
            Raises `TypeError` if threshold is not an integer.
            Default: `1000`.
        precision (int or None, optional): Number of digits printed for
            floating point numbers.
            Raises `TypeError` if precision is not an integer.
            Default: `8`.
        floatmode (str, optional): Controls how floating point
            numbers are interpreted.

            `"fixed:`: Always prints exactly `precision` digits.
            `"unique"`: Ignores precision, prints the number of
                digits necessary to uniquely specify each number.
            `"maxprec"`: Prints `precision` digits or fewer,
                if fewer will uniquely represent a number.
            `"maxprec_equal"`: Prints an equal number of digits
                for each number. This number is `precision` digits or fewer,
                if fewer will uniquely represent each number.
            Raises `ValueError` if floatmode is not one of
            `fixed`, `unique`, `maxprec`, or `maxprec_equal`.
            Default: "maxprec_equal"
        suppress (bool, optional): If `True,` numbers equal to zero
            in the current precision will print as zero.
            Default: `False`.
        nanstr (str, optional): String used to repesent nan.
            Raises `TypeError` if nanstr is not a string.
            Default: `"nan"`.
        infstr (str, optional): String used to represent infinity.
            Raises `TypeError` if infstr is not a string.
            Default: `"inf"`.
        sign (str, optional): Controls the sign of floating point
            numbers.
            `"-"`: Omit the sign of positive numbers.
            `"+"`: Always print the sign of positive numbers.
            `" "`: Always print a whitespace in place of the
                sign of positive numbers.
            Raises `ValueError` if sign is not one of
            `"-"`, `"+"`, or `" "`.
            Default: `"-"`.
        numpy (bool, optional): If `True,` then before other specified print
            options are set, a dictionary of Numpy's print options
            will be used to initialize dpctl's print options.
            Default: "False"
    """
    options = _options_dict(
        linewidth=linewidth,
        edgeitems=edgeitems,
        threshold=threshold,
        precision=precision,
        floatmode=floatmode,
        suppress=suppress,
        nanstr=nanstr,
        infstr=infstr,
        sign=sign,
        numpy=numpy,
    )
    _print_options.update(options)


def get_print_options():
    """
    get_print_options() -> dict

    Returns a copy of current options for printing
    ``dpctl.tensor.usm_ndarray`` class.

    Options:
        - "linewidth" : int, default 75
        - "edgeitems" : int, default 3
        - "threshold" : int, default 1000
        - "precision" : int, default 8
        - "floatmode" : str, default "maxprec_equal"
        - "suppress" : bool, default False
        - "nanstr" : str, default "nan"
        - "infstr" : str, default "inf"
        - "sign" : str, default "-"
    """
    return _print_options.copy()


@contextlib.contextmanager
def print_options(*args, **kwargs):
    """
    Context manager for print options.

    Set print options for the scope of a `with` block.
    `as` yields dictionary of print options.
    """
    options = dpt.get_print_options()
    try:
        dpt.set_print_options(*args, **kwargs)
        yield dpt.get_print_options()
    finally:
        dpt.set_print_options(**options)


def _nd_corners(x, edge_items, slices=()):
    axes_reduced = len(slices)
    if axes_reduced == x.ndim:
        return x[slices]

    if x.shape[axes_reduced] > 2 * edge_items:
        return dpt.concat(
            (
                _nd_corners(
                    x, edge_items, slices + (slice(None, edge_items, None),)
                ),
                _nd_corners(
                    x, edge_items, slices + (slice(-edge_items, None, None),)
                ),
            ),
            axis=axes_reduced,
        )
    else:
        return _nd_corners(x, edge_items, slices + (slice(None, None, None),))


def usm_ndarray_str(
    x,
    line_width=None,
    edge_items=None,
    threshold=None,
    precision=None,
    floatmode=None,
    suppress=None,
    sign=None,
    numpy=False,
    separator=" ",
    prefix="",
    suffix="",
):
    """
    usm_ndarray_str(x, line_width=None, edgeitems=None, threshold=None,
                    precision=None, floatmode=None, suppress=None,
                    sign=None, numpy=False, separator=" ", prefix="",
                    suffix="") -> str

    Returns a string representing the elements of a
    ``dpctl.tensor.usm_ndarray``.

    Args:
        x (usm_ndarray): Input array.
        line_width (int, optional): Number of characters printed per line.
            Raises `TypeError` if line_width is not an integer.
            Default: `75`.
        edgeitems (int, optional): Number of elements at the beginning and end
            when the printed array is abbreviated.
            Raises `TypeError` if edgeitems is not an integer.
            Default: `3`.
        threshold (int, optional): Number of elements that triggers array
            abbreviation.
            Raises `TypeError` if threshold is not an integer.
            Default: `1000`.
        precision (int or None, optional): Number of digits printed for
            floating point numbers.
            Raises `TypeError` if precision is not an integer.
            Default: `8`.
        floatmode (str, optional): Controls how floating point
            numbers are interpreted.

            `"fixed:`: Always prints exactly `precision` digits.
            `"unique"`: Ignores precision, prints the number of
                digits necessary to uniquely specify each number.
            `"maxprec"`: Prints `precision` digits or fewer,
                if fewer will uniquely represent a number.
            `"maxprec_equal"`: Prints an equal number of digits
                for each number. This number is `precision` digits or fewer,
                if fewer will uniquely represent each number.
            Raises `ValueError` if floatmode is not one of
            `fixed`, `unique`, `maxprec`, or `maxprec_equal`.
            Default: "maxprec_equal"
        suppress (bool, optional): If `True,` numbers equal to zero
            in the current precision will print as zero.
            Default: `False`.
        sign (str, optional): Controls the sign of floating point
            numbers.
            `"-"`: Omit the sign of positive numbers.
            `"+"`: Always print the sign of positive numbers.
            `" "`: Always print a whitespace in place of the
                sign of positive numbers.
            Raises `ValueError` if sign is not one of
            `"-"`, `"+"`, or `" "`.
            Default: `"-"`.
        numpy (bool, optional): If `True,` then before other specified print
            options are set, a dictionary of Numpy's print options
            will be used to initialize dpctl's print options.
            Default: "False"
        separator (str, optional): String inserted between elements of
            the array string.
            Default: " "
        prefix (str, optional): String used to determine spacing to the left
            of the array string.
            Default: ""
        suffix (str, optional): String that determines length of the last line
            of the array string.
            Default: ""
    """
    if not isinstance(x, dpt.usm_ndarray):
        raise TypeError(f"Expected dpctl.tensor.usm_ndarray, got {type(x)}")

    options = get_print_options()
    options.update(
        _options_dict(
            linewidth=line_width,
            edgeitems=edge_items,
            threshold=threshold,
            precision=precision,
            floatmode=floatmode,
            suppress=suppress,
            sign=sign,
            numpy=numpy,
        )
    )

    threshold = options["threshold"]
    edge_items = options["edgeitems"]

    if x.size > threshold:
        # need edge_items + 1 elements for np.array2string to abbreviate
        data = dpt.asnumpy(_nd_corners(x, edge_items + 1))
        options["threshold"] = 0
    else:
        data = dpt.asnumpy(x)
    with np.printoptions(**options):
        s = np.array2string(
            data, separator=separator, prefix=prefix, suffix=suffix
        )
    return s


def usm_ndarray_repr(
    x, line_width=None, precision=None, suppress=None, prefix="usm_ndarray"
):
    """
    usm_ndarray_repr(x, line_width=None, precision=None,
                     suppress=None, prefix="") -> str

        Returns a formatted string representing the elements
        of a ``dpctl.tensor.usm_ndarray`` and its data type,
        if not a default type.

    Args:
        x (usm_ndarray): Input array.
        line_width (int, optional): Number of characters printed per line.
            Raises `TypeError` if line_width is not an integer.
            Default: `75`.
        precision (int or None, optional): Number of digits printed for
            floating point numbers.
            Raises `TypeError` if precision is not an integer.
            Default: `8`.
        suppress (bool, optional): If `True,` numbers equal to zero
            in the current precision will print as zero.
            Default: `False`.
        prefix (str, optional): String inserted at the start of the array
            string.
            Default: ""
    """
    if not isinstance(x, dpt.usm_ndarray):
        raise TypeError(f"Expected dpctl.tensor.usm_ndarray, got {type(x)}")

    if line_width is None:
        line_width = _print_options["linewidth"]

    show_dtype = x.dtype not in [
        dpt.bool,
        dpt.int64,
        dpt.float64,
        dpt.complex128,
    ]

    prefix = prefix + "("
    suffix = ")"

    s = usm_ndarray_str(
        x,
        line_width=line_width,
        precision=precision,
        suppress=suppress,
        separator=", ",
        prefix=prefix,
        suffix=suffix,
    )

    if show_dtype:
        dtype_str = "dtype={}".format(x.dtype.name)
        bottom_len = len(s) - (s.rfind("\n") + 1)
        next_line = bottom_len + len(dtype_str) + 1 > line_width
        dtype_str = (
            ",\n" + " " * len(prefix) + dtype_str
            if next_line
            else ", " + dtype_str
        )
    else:
        dtype_str = ""

    return prefix + s + dtype_str + suffix



#  Copyright 2022 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# distutils: language = c++
# cython: language_level=3

# SYCL static imports for Cython

from . cimport _backend as dpctl_backend


cdef extern from "CL/sycl.hpp" namespace "sycl":
    cdef cppclass queue "sycl::queue":
        pass

    cdef cppclass device "sycl::device":
        pass

    cdef cppclass context "sycl::context":
        pass

    cdef cppclass event "sycl::event":
        pass

    cdef cppclass kernel "sycl::kernel":
        pass

    cdef cppclass executable_kernel_bundle \
        "sycl::kernel_bundle<sycl::bundle_state::executable>":
        pass

cdef extern from "syclinterface/dpctl_sycl_type_casters.hpp" \
    namespace "dpctl::syclinterface":
    # queue
    cdef dpctl_backend.DPCTLSyclQueueRef wrap_queue \
        "dpctl::syclinterface::wrap<sycl::queue>" (const queue *)
    cdef queue * unwrap_queue "dpctl::syclinterface::unwrap<sycl::queue>" (
        dpctl_backend.DPCTLSyclQueueRef)

    # device
    cdef dpctl_backend.DPCTLSyclDeviceRef wrap_device \
        "dpctl::syclinterface::wrap<sycl::device>" (const device *)
    cdef device * unwrap_device "dpctl::syclinterface::unwrap<sycl::device>" (
        dpctl_backend.DPCTLSyclDeviceRef)

    # context
    cdef dpctl_backend.DPCTLSyclContextRef wrap_context \
        "dpctl::syclinterface::wrap<sycl::context>" (const context *)
    cdef context * unwrap_context "dpctl::syclinterface::unwrap<sycl::context>" (
        dpctl_backend.DPCTLSyclContextRef)

    # event
    cdef dpctl_backend.DPCTLSyclEventRef wrap_event \
        "dpctl::syclinterface::wrap<sycl::event>" (const event *)
    cdef event * unwrap_event "dpctl::syclinterface::unwrap<sycl::event>" (
        dpctl_backend.DPCTLSyclEventRef)



#                      Data Parallel Control (dpctl)
#
# Copyright 2020-2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
import pytest
from helper import get_queue_or_skip, skip_if_dtype_not_supported
from numpy.testing import assert_array_equal

import dpctl.tensor as dpt
from dpctl.utils import ExecutionPlacementError

_all_dtypes = [
    "u1",
    "i1",
    "u2",
    "i2",
    "u4",
    "i4",
    "u8",
    "i8",
    "e",
    "f",
    "d",
    "F",
    "D",
]

_all_int_dtypes = ["u1", "i1", "u2", "i2", "u4", "i4", "u8", "i8"]


def test_basic_slice1():
    q = get_queue_or_skip()
    x = dpt.empty(10, dtype="u2", sycl_queue=q)
    y = x[0]
    assert isinstance(y, dpt.usm_ndarray)
    assert y.ndim == 0
    assert y.shape == tuple()
    assert y.strides == tuple()


def test_basic_slice2():
    q = get_queue_or_skip()
    x = dpt.empty(10, dtype="i2", sycl_queue=q)
    y = x[(0,)]
    assert isinstance(y, dpt.usm_ndarray)
    assert y.ndim == 0
    assert y.shape == tuple()
    assert y.strides == tuple()


def test_basic_slice3():
    q = get_queue_or_skip()
    x = dpt.empty(10, dtype="i2", sycl_queue=q)
    y = x[:]
    assert isinstance(y, dpt.usm_ndarray)
    assert y.ndim == x.ndim
    assert y.shape == x.shape
    assert y.strides == x.strides
    y = x[(slice(None, None, None),)]
    assert isinstance(y, dpt.usm_ndarray)
    assert y.ndim == x.ndim
    assert y.shape == x.shape
    assert y.strides == x.strides


def test_basic_slice4():
    q = get_queue_or_skip()
    n0, n1 = 5, 3
    x = dpt.empty((n0, n1), dtype="f4", sycl_queue=q)
    y = x[::-1]
    assert isinstance(y, dpt.usm_ndarray)
    assert y.shape == x.shape
    assert y.strides == (-x.strides[0], x.strides[1])
    actual_offset = y.__sycl_usm_array_interface__["offset"]
    assert actual_offset == (n0 - 1) * n1


def test_basic_slice5():
    q = get_queue_or_skip()
    n0, n1 = 5, 3
    x = dpt.empty((n0, n1), dtype="c8", sycl_queue=q)
    y = x[:, ::-1]
    assert isinstance(y, dpt.usm_ndarray)
    assert y.shape == x.shape
    assert y.strides == (x.strides[0], -x.strides[1])
    actual_offset = y.__sycl_usm_array_interface__["offset"]
    assert actual_offset == (n1 - 1)


def test_basic_slice6():
    q = get_queue_or_skip()
    i0, n0, n1 = 2, 4, 3
    x = dpt.empty((n0, n1), dtype="c8", sycl_queue=q)
    y = x[i0, ::-1]
    assert isinstance(y, dpt.usm_ndarray)
    assert y.shape == (x.shape[1],)
    assert y.strides == (-x.strides[1],)
    actual_offset = y.__sycl_usm_array_interface__["offset"]
    expected_offset = i0 * x.strides[0] + (n1 - 1) * x.strides[1]
    assert actual_offset == expected_offset


def test_basic_slice7():
    q = get_queue_or_skip()
    n0, n1, n2 = 5, 3, 2
    x = dpt.empty((n0, n1, n2), dtype="?", sycl_queue=q)
    y = x[..., ::-1]
    assert isinstance(y, dpt.usm_ndarray)
    assert y.shape == x.shape
    assert y.strides == (
        x.strides[0],
        x.strides[1],
        -x.strides[2],
    )
    actual_offset = y.__sycl_usm_array_interface__["offset"]
    expected_offset = (n2 - 1) * x.strides[2]
    assert actual_offset == expected_offset


def test_basic_slice8():
    q = get_queue_or_skip()
    n0, n1 = 3, 7
    x = dpt.empty((n0, n1), dtype="u1", sycl_queue=q)
    y = x[..., dpt.newaxis]
    assert isinstance(y, dpt.usm_ndarray)
    assert y.shape == (n0, n1, 1)
    assert y.strides == (n1, 1, 0)


def test_basic_slice9():
    q = get_queue_or_skip()
    n0, n1 = 3, 7
    x = dpt.empty((n0, n1), dtype="u8", sycl_queue=q)
    y = x[dpt.newaxis, ...]
    assert isinstance(y, dpt.usm_ndarray)
    assert y.shape == (1, n0, n1)
    assert y.strides == (0, n1, 1)


def test_basic_slice10():
    q = get_queue_or_skip()
    n0, n1, n2 = 3, 7, 5
    x = dpt.empty((n0, n1, n2), dtype="u1", sycl_queue=q)
    y = x[dpt.newaxis, ..., :]
    assert isinstance(y, dpt.usm_ndarray)
    assert y.shape == (1, n0, n1, n2)
    assert y.strides == (0, n1 * n2, n2, 1)


def _all_equal(it1, it2):
    return all(dpt.asnumpy(x) == dpt.asnumpy(y) for x, y in zip(it1, it2))


def test_advanced_slice1():
    q = get_queue_or_skip()
    ii = dpt.asarray([1, 2], sycl_queue=q)
    x = dpt.arange(10, dtype="i4", sycl_queue=q)
    y = x[ii]
    assert isinstance(y, dpt.usm_ndarray)
    assert y.shape == ii.shape
    assert y.strides == (1,)
    # FIXME, once usm_ndarray.__equal__ is implemented,
    # use of asnumpy should be removed
    assert _all_equal(
        (x[ii[k]] for k in range(ii.shape[0])),
        (y[k] for k in range(ii.shape[0])),
    )
    y = x[(ii,)]
    assert isinstance(y, dpt.usm_ndarray)
    assert y.shape == ii.shape
    assert y.strides == (1,)
    # FIXME, once usm_ndarray.__equal__ is implemented,
    # use of asnumpy should be removed
    assert _all_equal(
        (x[ii[k]] for k in range(ii.shape[0])),
        (y[k] for k in range(ii.shape[0])),
    )


def test_advanced_slice1_negative_strides():
    q = get_queue_or_skip()
    ii = dpt.asarray([0, 1], sycl_queue=q)
    x = dpt.flip(dpt.arange(5, dtype="i4", sycl_queue=q))
    y = x[ii]
    assert isinstance(y, dpt.usm_ndarray)
    assert y.shape == ii.shape
    assert y.strides == (1,)
    # FIXME, once usm_ndarray.__equal__ is implemented,
    # use of asnumpy should be removed
    assert _all_equal(
        (x[ii[k]] for k in range(ii.shape[0])),
        (y[k] for k in range(ii.shape[0])),
    )


def test_advanced_slice2():
    q = get_queue_or_skip()
    ii = dpt.asarray([1, 2], sycl_queue=q)
    x = dpt.arange(10, dtype="i4", sycl_queue=q)
    y = x[ii, dpt.newaxis]
    assert isinstance(y, dpt.usm_ndarray)
    assert y.shape == ii.shape + (1,)
    assert y.flags["C"]


def test_advanced_slice3():
    q = get_queue_or_skip()
    ii = dpt.asarray([1, 2], sycl_queue=q)
    x = dpt.arange(10, dtype="i4", sycl_queue=q)
    y = x[dpt.newaxis, ii]
    assert isinstance(y, dpt.usm_ndarray)
    assert y.shape == (1,) + ii.shape
    assert y.flags["C"]


def _make_3d(dt, q):
    return dpt.reshape(
        dpt.arange(3 * 3 * 3, dtype=dt, sycl_queue=q),
        (
            3,
            3,
            3,
        ),
    )


def test_advanced_slice4():
    q = get_queue_or_skip()
    ii = dpt.asarray([1, 2], sycl_queue=q)
    x = _make_3d("i4", q)
    y = x[ii, ii, ii]
    assert isinstance(y, dpt.usm_ndarray)
    assert y.shape == ii.shape
    assert _all_equal(
        (x[ii[k], ii[k], ii[k]] for k in range(ii.shape[0])),
        (y[k] for k in range(ii.shape[0])),
    )


def test_advanced_slice5():
    q = get_queue_or_skip()
    ii = dpt.asarray([1, 2], sycl_queue=q)
    x = _make_3d("i4", q)
    with pytest.raises(IndexError):
        x[ii, 0, ii]


def test_advanced_slice6():
    q = get_queue_or_skip()
    ii = dpt.asarray([1, 2], sycl_queue=q)
    x = _make_3d("i4", q)
    y = x[:, ii, ii]
    assert isinstance(y, dpt.usm_ndarray)
    assert y.shape == (
        x.shape[0],
        ii.shape[0],
    )
    assert _all_equal(
        (
            x[i, ii[k], ii[k]]
            for i in range(x.shape[0])
            for k in range(ii.shape[0])
        ),
        (y[i, k] for i in range(x.shape[0]) for k in range(ii.shape[0])),
    )


def test_advanced_slice7():
    q = get_queue_or_skip()
    mask = dpt.asarray(
        [
            [[True, True, False], [False, True, True], [True, False, True]],
            [[True, False, False], [False, False, True], [False, True, False]],
            [[True, True, True], [False, False, False], [False, False, True]],
        ],
        sycl_queue=q,
    )
    x = _make_3d("i2", q)
    y = x[mask]
    expected = [0, 1, 4, 5, 6, 8, 9, 14, 16, 18, 19, 20, 26]
    assert isinstance(y, dpt.usm_ndarray)
    assert y.shape == (len(expected),)
    assert all(dpt.asnumpy(y[k]) == expected[k] for k in range(len(expected)))


def test_advanced_slice8():
    q = get_queue_or_skip()
    mask = dpt.asarray(
        [[True, False, False], [False, True, False], [False, True, False]],
        sycl_queue=q,
    )
    x = _make_3d("u2", q)
    y = x[mask]
    expected = dpt.asarray(
        [[0, 1, 2], [12, 13, 14], [21, 22, 23]], sycl_queue=q
    )
    assert isinstance(y, dpt.usm_ndarray)
    assert y.shape == expected.shape
    assert (dpt.asnumpy(y) == dpt.asnumpy(expected)).all()


def test_advanced_slice9():
    q = get_queue_or_skip()
    mask = dpt.asarray(
        [[True, False, False], [False, True, False], [False, True, False]],
        sycl_queue=q,
    )
    x = _make_3d("u4", q)
    y = x[:, mask]
    expected = dpt.asarray([[0, 4, 7], [9, 13, 16], [18, 22, 25]], sycl_queue=q)
    assert isinstance(y, dpt.usm_ndarray)
    assert y.shape == expected.shape
    assert (dpt.asnumpy(y) == dpt.asnumpy(expected)).all()


def lin_id(i, j, k):
    """global_linear_id for (3,3,3) range traversed in C-contiguous order"""
    return 9 * i + 3 * j + k


def test_advanced_slice10():
    q = get_queue_or_skip()
    x = _make_3d("u8", q)
    i0 = dpt.asarray([0, 1, 1], device=x.device)
    i1 = dpt.asarray([1, 1, 2], device=x.device)
    i2 = dpt.asarray([2, 0, 1], device=x.device)
    y = x[i0, i1, i2]
    res_expected = dpt.asarray(
        [
            lin_id(0, 1, 2),
            lin_id(1, 1, 0),
            lin_id(1, 2, 1),
        ],
        sycl_queue=q,
    )
    assert isinstance(y, dpt.usm_ndarray)
    assert y.shape == res_expected.shape
    assert (dpt.asnumpy(y) == dpt.asnumpy(res_expected)).all()


def test_advanced_slice11():
    q = get_queue_or_skip()
    x = _make_3d("u8", q)
    i0 = dpt.asarray([0, 1, 1], device=x.device)
    i2 = dpt.asarray([2, 0, 1], device=x.device)
    with pytest.raises(IndexError):
        x[i0, :, i2]


def test_advanced_slice12():
    q = get_queue_or_skip()
    x = _make_3d("u8", q)
    i1 = dpt.asarray([1, 1, 2], device=x.device)
    i2 = dpt.asarray([2, 0, 1], device=x.device)
    y = x[:, dpt.newaxis, i1, i2, dpt.newaxis]
    res_expected = dpt.asarray(
        [
            [[[lin_id(0, 1, 2)], [lin_id(0, 1, 0)], [lin_id(0, 2, 1)]]],
            [[[lin_id(1, 1, 2)], [lin_id(1, 1, 0)], [lin_id(1, 2, 1)]]],
            [[[lin_id(2, 1, 2)], [lin_id(2, 1, 0)], [lin_id(2, 2, 1)]]],
        ],
        sycl_queue=q,
    )
    assert isinstance(y, dpt.usm_ndarray)
    assert y.shape == res_expected.shape
    assert (dpt.asnumpy(y) == dpt.asnumpy(res_expected)).all()


def test_advanced_slice13():
    q = get_queue_or_skip()
    x = _make_3d("u8", q)
    i1 = dpt.asarray([[1], [2]], device=x.device)
    i2 = dpt.asarray([[0, 1]], device=x.device)
    y = x[i1, i2, 0]
    expected = dpt.asarray(
        [
            [lin_id(1, 0, 0), lin_id(1, 1, 0)],
            [lin_id(2, 0, 0), lin_id(2, 1, 0)],
        ],
        device=x.device,
    )
    assert isinstance(y, dpt.usm_ndarray)
    assert y.shape == expected.shape
    assert (dpt.asnumpy(y) == dpt.asnumpy(expected)).all()


def test_integer_indexing_1d():
    get_queue_or_skip()
    x = dpt.arange(10, dtype="i4")
    ind_1d = dpt.asarray([7, 3, 1], dtype="u2")
    ind_2d = dpt.asarray([[2, 3, 4], [3, 4, 5], [5, 6, 7]], dtype="i4")

    y1 = x[ind_1d]
    assert y1.shape == ind_1d.shape
    y2 = x[ind_2d]
    assert y2.shape == ind_2d.shape
    assert (dpt.asnumpy(y1) == np.array([7, 3, 1], dtype="i4")).all()
    assert (
        dpt.asnumpy(y2)
        == np.array([[2, 3, 4], [3, 4, 5], [5, 6, 7]], dtype="i4")
    ).all()


def test_integer_indexing_2d():
    get_queue_or_skip()
    n0, n1 = 5, 7
    x = dpt.reshape(
        dpt.arange(n0 * n1, dtype="i4"),
        (
            n0,
            n1,
        ),
    )
    ind0 = dpt.arange(n0)
    ind1 = dpt.arange(n1)

    y = x[ind0[:2, dpt.newaxis], ind1[dpt.newaxis, -2:]]
    assert y.dtype == x.dtype
    assert (dpt.asnumpy(y) == np.array([[5, 6], [12, 13]])).all()


def test_integer_strided_indexing():
    get_queue_or_skip()
    n0, n1 = 5, 7
    x = dpt.reshape(
        dpt.arange(2 * n0 * n1, dtype="i4"),
        (
            2 * n0,
            n1,
        ),
    )
    ind0 = dpt.arange(n0)
    ind1 = dpt.arange(n1)

    z = x[::-2, :]
    y = z[ind0[:2, dpt.newaxis], ind1[dpt.newaxis, -2:]]
    assert y.dtype == x.dtype
    zc = dpt.copy(z, order="C")
    yc = zc[ind0[:2, dpt.newaxis], ind1[dpt.newaxis, -2:]]
    assert (dpt.asnumpy(y) == dpt.asnumpy(yc)).all()


@pytest.mark.parametrize(
    "data_dt",
    _all_dtypes,
)
@pytest.mark.parametrize(
    "ind_dt",
    _all_int_dtypes,
)
def test_take_basic(data_dt, ind_dt):
    q = get_queue_or_skip()
    skip_if_dtype_not_supported(data_dt, q)

    x = dpt.arange(10, dtype=data_dt)
    ind = dpt.arange(2, 5, dtype=ind_dt)
    y = dpt.take(x, ind)
    assert y.dtype == x.dtype
    assert (dpt.asnumpy(y) == np.arange(2, 5, dtype=data_dt)).all()


@pytest.mark.parametrize(
    "data_dt",
    _all_dtypes,
)
@pytest.mark.parametrize(
    "ind_dt",
    _all_int_dtypes,
)
def test_put_basic(data_dt, ind_dt):
    q = get_queue_or_skip()
    skip_if_dtype_not_supported(data_dt, q)

    x = dpt.arange(10, dtype=data_dt)
    ind = dpt.arange(2, 5, dtype=ind_dt)
    val = dpt.ones(3, dtype=data_dt)
    dpt.put(x, ind, val)
    assert (
        dpt.asnumpy(x)
        == np.array([0, 1, 1, 1, 1, 5, 6, 7, 8, 9], dtype=data_dt)
    ).all()


def test_take_basic_axis():
    get_queue_or_skip()

    n0, n1 = 5, 7
    x = dpt.reshape(
        dpt.arange(n0 * n1, dtype="i4"),
        (
            n0,
            n1,
        ),
    )
    ind = dpt.arange(2, 4)
    y0 = dpt.take(x, ind, axis=0)
    y1 = dpt.take(x, ind, axis=1)
    assert y0.shape == (2, n1)
    assert y1.shape == (n0, 2)


def test_put_basic_axis():
    get_queue_or_skip()

    n0, n1 = 5, 7
    x = dpt.reshape(
        dpt.arange(n0 * n1, dtype="i4"),
        (
            n0,
            n1,
        ),
    )
    ind = dpt.arange(2, 4)
    v0 = dpt.zeros((2, n1), dtype=x.dtype)
    v1 = dpt.zeros((n0, 2), dtype=x.dtype)
    dpt.put(x, ind, v0, axis=0)
    dpt.put(x, ind, v1, axis=1)
    expected = np.arange(n0 * n1, dtype="i4").reshape((n0, n1))
    expected[[2, 3], :] = 0
    expected[:, [2, 3]] = 0
    assert (expected == dpt.asnumpy(x)).all()


@pytest.mark.parametrize("data_dt", _all_dtypes)
def test_put_0d_val(data_dt):
    q = get_queue_or_skip()
    skip_if_dtype_not_supported(data_dt, q)

    x = dpt.arange(5, dtype=data_dt, sycl_queue=q)
    ind = dpt.asarray([0], dtype=np.intp, sycl_queue=q)
    val = dpt.asarray(2, dtype=x.dtype, sycl_queue=q)
    x[ind] = val
    assert_array_equal(np.asarray(2, dtype=data_dt), dpt.asnumpy(x[0]))

    x = dpt.asarray(5, dtype=data_dt, sycl_queue=q)
    dpt.put(x, ind, val)
    assert_array_equal(np.asarray(2, dtype=data_dt), dpt.asnumpy(x))


@pytest.mark.parametrize(
    "data_dt",
    _all_dtypes,
)
def test_take_0d_data(data_dt):
    q = get_queue_or_skip()
    skip_if_dtype_not_supported(data_dt, q)

    x = dpt.asarray(0, dtype=data_dt, sycl_queue=q)
    ind = dpt.arange(5, dtype=np.intp, sycl_queue=q)

    y = dpt.take(x, ind)
    assert (
        dpt.asnumpy(y)
        == np.broadcast_to(np.asarray(0, dtype=data_dt), ind.shape)
    ).all()


@pytest.mark.parametrize(
    "data_dt",
    _all_dtypes,
)
def test_put_0d_data(data_dt):
    q = get_queue_or_skip()
    skip_if_dtype_not_supported(data_dt, q)

    x = dpt.asarray(0, dtype=data_dt, sycl_queue=q)
    ind = dpt.arange(5, dtype=np.intp, sycl_queue=q)
    val = dpt.asarray(2, dtype=data_dt, sycl_queue=q)

    dpt.put(x, ind, val, axis=0)
    assert (
        dpt.asnumpy(x)
        == np.broadcast_to(np.asarray(2, dtype=data_dt), ind.shape)
    ).all()


@pytest.mark.parametrize(
    "ind_dt",
    _all_int_dtypes,
)
def test_indexing_0d_ind(ind_dt):
    q = get_queue_or_skip()

    x = dpt.arange(5, dtype="i4", sycl_queue=q)
    ind = dpt.asarray(3, dtype=ind_dt, sycl_queue=q)

    y = x[ind]
    assert dpt.asnumpy(x[3]) == dpt.asnumpy(y)


@pytest.mark.parametrize(
    "ind_dt",
    _all_int_dtypes,
)
def test_put_0d_ind(ind_dt):
    q = get_queue_or_skip()

    x = dpt.arange(5, dtype="i4", sycl_queue=q)
    ind = dpt.asarray(3, dtype=ind_dt, sycl_queue=q)
    val = dpt.asarray(5, dtype=x.dtype, sycl_queue=q)

    x[ind] = val
    assert dpt.asnumpy(x[3]) == dpt.asnumpy(val)


@pytest.mark.parametrize(
    "data_dt",
    _all_dtypes,
)
def test_take_strided_1d_source(data_dt):
    q = get_queue_or_skip()
    skip_if_dtype_not_supported(data_dt, q)

    x = dpt.arange(27, dtype=data_dt, sycl_queue=q)
    ind = dpt.arange(4, 9, dtype=np.intp, sycl_queue=q)

    x_np = dpt.asnumpy(x)
    ind_np = dpt.asnumpy(ind)

    for s in (
        slice(None, None, 2),
        slice(None, None, -2),
    ):
        assert_array_equal(
            np.take(x_np[s], ind_np, axis=0),
            dpt.asnumpy(dpt.take(x[s], ind, axis=0)),
        )

    # 0-strided
    x = dpt.usm_ndarray(
        (27,),
        dtype=data_dt,
        strides=(0,),
        buffer_ctor_kwargs={"queue": q},
    )
    x[0] = x_np[0]
    assert_array_equal(
        np.broadcast_to(x_np[0], ind.shape),
        dpt.asnumpy(dpt.take(x, ind, axis=0)),
    )


@pytest.mark.parametrize(
    "data_dt",
    _all_dtypes,
)
@pytest.mark.parametrize("order", ["C", "F"])
def test_take_strided(data_dt, order):
    q = get_queue_or_skip()
    skip_if_dtype_not_supported(data_dt, q)

    x = dpt.reshape(_make_3d(data_dt, q), (9, 3), order=order)
    ind = dpt.arange(2, dtype=np.intp, sycl_queue=q)

    x_np = dpt.asnumpy(x)
    ind_np = dpt.asnumpy(ind)

    for s in (
        slice(None, None, 2),
        slice(None, None, -2),
    ):
        for sgn in (-1, 1):
            xs = x[s, ::sgn]
            xs_np = x_np[s, ::sgn]
            assert_array_equal(
                np.take(xs_np, ind_np, axis=0),
                dpt.asnumpy(dpt.take(xs, ind, axis=0)),
            )
            assert_array_equal(
                np.take(xs_np, ind_np, axis=1),
                dpt.asnumpy(dpt.take(xs, ind, axis=1)),
            )


@pytest.mark.parametrize(
    "ind_dt",
    _all_int_dtypes,
)
def test_take_strided_1d_indices(ind_dt):
    q = get_queue_or_skip()

    x = dpt.arange(27, dtype="i4", sycl_queue=q)
    ind = dpt.arange(12, 24, dtype=ind_dt, sycl_queue=q)

    x_np = dpt.asnumpy(x)
    ind_np = dpt.asnumpy(ind).astype(np.intp)

    for s in (
        slice(None, None, 2),
        slice(None, None, -2),
    ):
        assert_array_equal(
            np.take(x_np, ind_np[s], axis=0),
            dpt.asnumpy(dpt.take(x, ind[s], axis=0)),
        )

    # 0-strided
    ind = dpt.usm_ndarray(
        (12,),
        dtype=ind_dt,
        strides=(0,),
        buffer_ctor_kwargs={"queue": q},
    )
    ind[0] = ind_np[0]
    assert_array_equal(
        np.broadcast_to(x_np[ind_np[0]], ind.shape),
        dpt.asnumpy(dpt.take(x, ind, axis=0)),
    )


@pytest.mark.parametrize(
    "ind_dt",
    _all_int_dtypes,
)
@pytest.mark.parametrize("order", ["C", "F"])
def test_take_strided_indices(ind_dt, order):
    q = get_queue_or_skip()

    x = dpt.arange(27, dtype="i4", sycl_queue=q)
    ind = dpt.reshape(
        dpt.arange(12, 24, dtype=ind_dt, sycl_queue=q), (4, 3), order=order
    )

    x_np = dpt.asnumpy(x)
    ind_np = dpt.asnumpy(ind).astype(np.intp)

    for s in (
        slice(None, None, 2),
        slice(None, None, -2),
    ):
        for sgn in [-1, 1]:
            inds = ind[s, ::sgn]
            inds_np = ind_np[s, ::sgn]
            assert_array_equal(
                np.take(x_np, inds_np, axis=0),
                dpt.asnumpy(x[inds]),
            )


@pytest.mark.parametrize(
    "data_dt",
    _all_dtypes,
)
@pytest.mark.parametrize("order", ["C", "F"])
def test_put_strided_1d_destination(data_dt, order):
    q = get_queue_or_skip()
    skip_if_dtype_not_supported(data_dt, q)

    x = dpt.arange(27, dtype=data_dt, sycl_queue=q)
    ind = dpt.arange(4, 9, dtype=np.intp, sycl_queue=q)
    val = dpt.asarray(9, dtype=x.dtype, sycl_queue=q)

    x_np = dpt.asnumpy(x)
    ind_np = dpt.asnumpy(ind)
    val_np = dpt.asnumpy(val)

    for s in (
        slice(None, None, 2),
        slice(None, None, -2),
    ):
        x_np1 = x_np.copy()
        x_np1[s][ind_np] = val_np

        x1 = dpt.copy(x)
        dpt.put(x1[s], ind, val, axis=0)

        assert_array_equal(x_np1, dpt.asnumpy(x1))


@pytest.mark.parametrize(
    "data_dt",
    _all_dtypes,
)
@pytest.mark.parametrize("order", ["C", "F"])
def test_put_strided_destination(data_dt, order):
    q = get_queue_or_skip()
    skip_if_dtype_not_supported(data_dt, q)

    x = dpt.reshape(_make_3d(data_dt, q), (9, 3), order=order)
    ind = dpt.arange(2, dtype=np.intp, sycl_queue=q)
    val = dpt.asarray(9, dtype=x.dtype, sycl_queue=q)

    x_np = dpt.asnumpy(x)
    ind_np = dpt.asnumpy(ind)
    val_np = dpt.asnumpy(val)

    for s in (
        slice(None, None, 2),
        slice(None, None, -2),
    ):
        for sgn in [-1, 1]:
            xs = x[s, ::sgn]
            xs_np = x_np[s, ::sgn]

            x_np1 = xs_np.copy()
            x_np1[ind_np] = val_np

            x1 = dpt.copy(xs)
            dpt.put(x1, ind, val, axis=0)
            assert_array_equal(x_np1, dpt.asnumpy(x1))

            x_np1 = xs_np.copy()
            x_np1[:, ind_np] = val_np

            x1 = dpt.copy(xs)
            dpt.put(x1, ind, val, axis=1)
            assert_array_equal(x_np1, dpt.asnumpy(x1))

            x_np1 = xs_np.copy()
            x_np1[ind_np, ind_np] = val_np

            x1 = dpt.copy(xs)
            x1[ind, ind] = val
            assert_array_equal(x_np1, dpt.asnumpy(x1))


@pytest.mark.parametrize(
    "ind_dt",
    _all_int_dtypes,
)
def test_put_strided_1d_indices(ind_dt):
    q = get_queue_or_skip()

    x = dpt.arange(27, dtype="i4", sycl_queue=q)
    ind = dpt.arange(12, 24, dtype=ind_dt, sycl_queue=q)
    val = dpt.asarray(-1, dtype=x.dtype, sycl_queue=q)

    x_np = dpt.asnumpy(x)
    ind_np = dpt.asnumpy(ind).astype(np.intp)
    val_np = dpt.asnumpy(val)

    for s in (
        slice(None, None, 2),
        slice(None, None, -2),
    ):
        x_copy = dpt.copy(x)
        dpt.put(x_copy, ind[s], val, axis=0)

        x_np_copy = x_np.copy()
        x_np_copy[ind_np[s]] = val_np

        assert_array_equal(x_np_copy, dpt.asnumpy(x_copy))


@pytest.mark.parametrize(
    "ind_dt",
    _all_int_dtypes,
)
@pytest.mark.parametrize("order", ["C", "F"])
def test_put_strided_indices(ind_dt, order):
    q = get_queue_or_skip()

    x = dpt.arange(27, dtype="i4", sycl_queue=q)
    ind = dpt.reshape(
        dpt.arange(12, 24, dtype=ind_dt, sycl_queue=q), (4, 3), order=order
    )
    val = dpt.asarray(-1, sycl_queue=q, dtype=x.dtype)

    x_np = dpt.asnumpy(x)
    ind_np = dpt.asnumpy(ind).astype(np.intp)
    val_np = dpt.asnumpy(val)

    for s in (
        slice(None, None, 2),
        slice(None, None, -2),
    ):
        for sgn in [-1, 1]:
            inds = ind[s, ::sgn]
            inds_np = ind_np[s, ::sgn]

            x_copy = dpt.copy(x)
            x_copy[inds] = val

            x_np_copy = x_np.copy()
            x_np_copy[inds_np] = val_np

            assert_array_equal(x_np_copy, dpt.asnumpy(x_copy))


def test_take_arg_validation():
    q = get_queue_or_skip()

    x = dpt.arange(4, dtype="i4", sycl_queue=q)
    ind0 = dpt.arange(4, dtype=np.intp, sycl_queue=q)
    ind1 = dpt.arange(2.0, dtype="f", sycl_queue=q)

    with pytest.raises(TypeError):
        dpt.take(dict(), ind0, axis=0)
    with pytest.raises(TypeError):
        dpt.take(x, dict(), axis=0)
    with pytest.raises(TypeError):
        x[[]]
    with pytest.raises(IndexError):
        dpt.take(x, ind1, axis=0)
    with pytest.raises(IndexError):
        x[ind1]

    with pytest.raises(ValueError):
        dpt.take(dpt.reshape(x, (2, 2)), ind0)
    with pytest.raises(ValueError):
        dpt.take(x, ind0, mode=0)
    with pytest.raises(ValueError):
        dpt.take(dpt.reshape(x, (2, 2)), ind0, axis=None)
    with pytest.raises(ValueError):
        dpt.take(x, dpt.reshape(ind0, (2, 2)))


def test_put_arg_validation():
    q = get_queue_or_skip()

    x = dpt.arange(4, dtype="i4", sycl_queue=q)
    ind0 = dpt.arange(4, dtype=np.intp, sycl_queue=q)
    ind1 = dpt.arange(2.0, dtype="f", sycl_queue=q)
    val = dpt.asarray(2, x.dtype, sycl_queue=q)

    with pytest.raises(TypeError):
        dpt.put(dict(), ind0, val, axis=0)
    with pytest.raises(TypeError):
        dpt.put(x, dict(), val, axis=0)
    with pytest.raises(TypeError):
        x[[]] = val
    with pytest.raises(IndexError):
        dpt.put(x, ind1, val, axis=0)
    with pytest.raises(IndexError):
        x[ind1] = val
    with pytest.raises(TypeError):
        dpt.put(x, ind0, dict(), axis=0)
    with pytest.raises(TypeError):
        x[ind0] = dict()

    with pytest.raises(ValueError):
        dpt.put(x, ind0, val, mode=0)
    with pytest.raises(ValueError):
        dpt.put(x, dpt.reshape(ind0, (2, 2)), val)


def test_advanced_indexing_compute_follows_data():
    q1 = get_queue_or_skip()
    q2 = get_queue_or_skip()

    x = dpt.arange(4, sycl_queue=q1)
    ind0 = dpt.asarray([0], sycl_queue=q1)
    ind1 = dpt.asarray([0], sycl_queue=q2)
    val0 = dpt.asarray(2, dtype=x.dtype, sycl_queue=q1)
    val1 = dpt.asarray(2, dtype=x.dtype, sycl_queue=q2)

    with pytest.raises(ExecutionPlacementError):
        dpt.take(x, ind1, axis=0)
    with pytest.raises(ExecutionPlacementError):
        x[ind1]
    with pytest.raises(ExecutionPlacementError):
        dpt.put(x, ind1, val0, axis=0)
    with pytest.raises(ExecutionPlacementError):
        x[ind1] = val0
    with pytest.raises(ExecutionPlacementError):
        dpt.put(x, ind0, val1, axis=0)
    with pytest.raises(ExecutionPlacementError):
        x[ind0] = val1


def test_extract_all_1d():
    get_queue_or_skip()
    x = dpt.arange(30, dtype="i4")
    sel = dpt.ones(30, dtype="?")
    sel[::2] = False

    res = x[sel]
    expected_res = dpt.asnumpy(x)[dpt.asnumpy(sel)]
    assert (dpt.asnumpy(res) == expected_res).all()

    res2 = dpt.extract(sel, x)
    assert (dpt.asnumpy(res2) == expected_res).all()


def test_extract_all_2d():
    get_queue_or_skip()
    x = dpt.reshape(dpt.arange(30, dtype="i4"), (5, 6))
    sel = dpt.ones(30, dtype="?")
    sel[::2] = False
    sel = dpt.reshape(sel, x.shape)

    res = x[sel]
    expected_res = dpt.asnumpy(x)[dpt.asnumpy(sel)]
    assert (dpt.asnumpy(res) == expected_res).all()

    res2 = dpt.extract(sel, x)
    assert (dpt.asnumpy(res2) == expected_res).all()


def test_extract_2D_axis0():
    get_queue_or_skip()
    x = dpt.reshape(dpt.arange(30, dtype="i4"), (5, 6))
    sel = dpt.ones(x.shape[0], dtype="?")
    sel[::2] = False

    res = x[sel]
    expected_res = dpt.asnumpy(x)[dpt.asnumpy(sel)]
    assert (dpt.asnumpy(res) == expected_res).all()


def test_extract_2D_axis1():
    get_queue_or_skip()
    x = dpt.reshape(dpt.arange(30, dtype="i4"), (5, 6))
    sel = dpt.ones(x.shape[1], dtype="?")
    sel[::2] = False

    res = x[:, sel]
    expected = dpt.asnumpy(x)[:, dpt.asnumpy(sel)]
    assert (dpt.asnumpy(res) == expected).all()


def test_extract_begin():
    get_queue_or_skip()
    x = dpt.reshape(dpt.arange(3 * 3 * 4 * 4, dtype="i2"), (3, 4, 3, 4))
    y = dpt.permute_dims(x, (2, 0, 3, 1))
    sel = dpt.zeros((3, 3), dtype="?")
    sel[0, 0] = True
    sel[1, 1] = True
    z = y[sel]
    expected = dpt.asnumpy(y)[[0, 1], [0, 1]]
    assert (dpt.asnumpy(z) == expected).all()


def test_extract_end():
    get_queue_or_skip()
    x = dpt.reshape(dpt.arange(3 * 3 * 4 * 4, dtype="i2"), (3, 4, 3, 4))
    y = dpt.permute_dims(x, (2, 0, 3, 1))
    sel = dpt.zeros((4, 4), dtype="?")
    sel[0, 0] = True
    z = y[..., sel]
    expected = dpt.asnumpy(y)[..., [0], [0]]
    assert (dpt.asnumpy(z) == expected).all()


def test_extract_middle():
    get_queue_or_skip()
    x = dpt.reshape(dpt.arange(3 * 3 * 4 * 4, dtype="i2"), (3, 4, 3, 4))
    y = dpt.permute_dims(x, (2, 0, 3, 1))
    sel = dpt.zeros((3, 4), dtype="?")
    sel[0, 0] = True
    z = y[:, sel]
    expected = dpt.asnumpy(y)[:, [0], [0], :]
    assert (dpt.asnumpy(z) == expected).all()


def test_extract_empty_result():
    get_queue_or_skip()
    x = dpt.reshape(dpt.arange(3 * 3 * 4 * 4, dtype="i2"), (3, 4, 3, 4))
    y = dpt.permute_dims(x, (2, 0, 3, 1))
    sel = dpt.zeros((3, 4), dtype="?")
    z = y[:, sel]
    assert z.shape == (
        y.shape[0],
        0,
        y.shape[3],
    )


def test_place_all_1d():
    get_queue_or_skip()
    x = dpt.arange(10, dtype="i2")
    sel = dpt.zeros(10, dtype="?")
    sel[0::2] = True
    val = dpt.zeros(5, dtype=x.dtype)
    x[sel] = val
    assert (dpt.asnumpy(x) == np.array([0, 1, 0, 3, 0, 5, 0, 7, 0, 9])).all()
    dpt.place(x, sel, dpt.asarray([2]))
    assert (dpt.asnumpy(x) == np.array([2, 1, 2, 3, 2, 5, 2, 7, 2, 9])).all()


def test_place_2d_axis0():
    get_queue_or_skip()
    x = dpt.reshape(dpt.arange(12, dtype="i2"), (3, 4))
    sel = dpt.asarray([True, False, True])
    val = dpt.zeros((2, 4), dtype=x.dtype)
    x[sel] = val
    expected_x = np.stack(
        (
            np.zeros(4, dtype="i2"),
            np.arange(4, 8, dtype="i2"),
            np.zeros(4, dtype="i2"),
        )
    )
    assert (dpt.asnumpy(x) == expected_x).all()


def test_place_2d_axis1():
    get_queue_or_skip()
    x = dpt.reshape(dpt.arange(12, dtype="i2"), (3, 4))
    sel = dpt.asarray([True, False, True, False])
    val = dpt.zeros((3, 2), dtype=x.dtype)
    x[:, sel] = val
    expected_x = np.array(
        [[0, 1, 0, 3], [0, 5, 0, 7], [0, 9, 0, 11]], dtype="i2"
    )
    assert (dpt.asnumpy(x) == expected_x).all()


def test_place_2d_axis1_scalar():
    get_queue_or_skip()
    x = dpt.reshape(dpt.arange(12, dtype="i2"), (3, 4))
    sel = dpt.asarray([True, False, True, False])
    val = dpt.zeros(tuple(), dtype=x.dtype)
    x[:, sel] = val
    expected_x = np.array(
        [[0, 1, 0, 3], [0, 5, 0, 7], [0, 9, 0, 11]], dtype="i2"
    )
    assert (dpt.asnumpy(x) == expected_x).all()


def test_place_all_slices():
    get_queue_or_skip()
    x = dpt.reshape(dpt.arange(12, dtype="i2"), (3, 4))
    sel = dpt.asarray(
        [
            [False, True, True, False],
            [True, True, False, False],
            [False, False, True, True],
        ],
        dtype="?",
    )
    y = dpt.ones_like(x)
    y[sel] = x[sel]


def test_place_some_slices_begin():
    get_queue_or_skip()
    x = dpt.reshape(dpt.arange(3 * 3 * 4 * 4, dtype="i2"), (3, 4, 3, 4))
    y = dpt.permute_dims(x, (2, 0, 3, 1))
    sel = dpt.zeros((3, 3), dtype="?")
    sel[0, 0] = True
    sel[1, 1] = True
    z = y[sel]
    w = dpt.zeros_like(y)
    w[sel] = z


def test_place_some_slices_mid():
    get_queue_or_skip()
    x = dpt.reshape(dpt.arange(3 * 3 * 4 * 4, dtype="i2"), (3, 4, 3, 4))
    y = dpt.permute_dims(x, (2, 0, 3, 1))
    sel = dpt.zeros((3, 4), dtype="?")
    sel[0, 0] = True
    sel[1, 1] = True
    z = y[:, sel]
    w = dpt.zeros_like(y)
    w[:, sel] = z


def test_place_some_slices_end():
    get_queue_or_skip()
    x = dpt.reshape(dpt.arange(3 * 3 * 4 * 4, dtype="i2"), (3, 4, 3, 4))
    y = dpt.permute_dims(x, (2, 0, 3, 1))
    sel = dpt.zeros((4, 4), dtype="?")
    sel[0, 0] = True
    sel[1, 1] = True
    z = y[:, :, sel]
    w = dpt.zeros_like(y)
    w[:, :, sel] = z


def test_place_cycling():
    get_queue_or_skip()
    x = dpt.zeros(10, dtype="f4")
    y = dpt.asarray([2, 3])
    sel = dpt.ones(x.size, dtype="?")
    dpt.place(x, sel, y)
    expected = np.array(
        [
            2,
            3,
        ]
        * 5,
        dtype=x.dtype,
    )
    assert (dpt.asnumpy(x) == expected).all()


def test_place_subset():
    get_queue_or_skip()
    x = dpt.zeros(10, dtype="f4")
    y = dpt.ones_like(x)
    sel = dpt.ones(x.size, dtype="?")
    sel[::2] = False
    dpt.place(x, sel, y)
    expected = np.array([0, 1, 0, 1, 0, 1, 0, 1, 0, 1], dtype=x.dtype)
    assert (dpt.asnumpy(x) == expected).all()


def test_nonzero():
    get_queue_or_skip()
    x = dpt.concat((dpt.zeros(3), dpt.ones(4), dpt.zeros(3)))
    (i,) = dpt.nonzero(x)
    assert (dpt.asnumpy(i) == np.array([3, 4, 5, 6])).all()


def test_assign_scalar():
    get_queue_or_skip()
    x = dpt.arange(-5, 5, dtype="i8")
    cond = dpt.asarray(
        [True, True, True, True, True, False, False, False, False, False]
    )
    x[cond] = 0  # no error expected
    x[dpt.nonzero(cond)] = -1
    expected = np.array([-1, -1, -1, -1, -1, 0, 1, 2, 3, 4], dtype=x.dtype)
    assert (dpt.asnumpy(x) == expected).all()


def test_nonzero_large():
    get_queue_or_skip()
    m = dpt.full((60, 80), True)
    assert m[m].size == m.size

    m = dpt.full((30, 60, 80), True)
    assert m[m].size == m.size



#                      Data Parallel Control (dpctl)
#
# Copyright 2020-2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper module for dpctl/tests
"""

from ._helper import (
    create_invalid_capsule,
    get_queue_or_skip,
    has_cpu,
    has_gpu,
    has_sycl_platforms,
    skip_if_dtype_not_supported,
)

__all__ = [
    "create_invalid_capsule",
    "has_cpu",
    "has_gpu",
    "has_sycl_platforms",
    "get_queue_or_skip",
    "skip_if_dtype_not_supported",
]



#                      Data Parallel Control (dpctl)
#
# Copyright 2020-2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Defines unit test cases for the SyclDevice class.
"""

import pytest

import dpctl
from dpctl import SyclDeviceCreationError


def test_standard_selectors(device_selector, check):
    """Tests if the standard SYCL device_selectors are able to select a
    device.
    """
    try:
        device = device_selector()
        check(device)
    except dpctl.SyclDeviceCreationError:
        pytest.skip()


def test_current_device(check):
    """Test is the device for the current queue is valid."""
    try:
        q = dpctl.get_current_queue()
    except Exception:
        pytest.fail("Encountered an exception inside get_current_queue().")
    device = q.get_sycl_device()
    check(device)


def test_valid_filter_selectors(valid_filter, check):
    """
    Tests if we can create a SyclDevice using a supported filter selector
    string.
    """
    device = None
    try:
        device = dpctl.SyclDevice(valid_filter)
    except SyclDeviceCreationError:
        pytest.skip("Failed to create device with supported filter")
    check(device)


def test_invalid_filter_selectors(invalid_filter):
    """
    An invalid filter string should always be caught and a ValueError raised.
    """
    exc = (
        SyclDeviceCreationError
        if isinstance(invalid_filter, str)
        else ValueError
    )
    with pytest.raises(exc):
        dpctl.SyclDevice(invalid_filter)


def test_filter_string(valid_filter):
    """
    Test that filter_string reconstructs the same device.
    """
    device = None
    try:
        device = dpctl.SyclDevice(valid_filter)
    except SyclDeviceCreationError:
        pytest.skip("Failed to create device with supported filter")
    dev_id = device.filter_string
    assert (
        dpctl.SyclDevice(dev_id) == device
    ), "Reconstructed device is different, ({}, {})".format(
        valid_filter, dev_id
    )


def test_filter_string_property():
    """
    Test that filter_string reconstructs the same device.
    """
    devices = dpctl.get_devices()
    for d in devices:
        if d.default_selector_score >= 0:
            dev_id = d.filter_string
            d_r = dpctl.SyclDevice(dev_id)
            assert d == d_r
            assert hash(d) == hash(d_r)


def test_filter_string_method():
    """
    Test that filter_string reconstructs the same device.
    """
    devices = dpctl.get_devices()
    for d in devices:
        for be in [True, False]:
            for dt in [True, False]:
                if d.default_selector_score >= 0:
                    dev_id = d.get_filter_string(
                        include_backend=be, include_device_type=dt
                    )
                    d_r = dpctl.SyclDevice(dev_id)
                    assert d == d_r, "Failed "
                    assert hash(d) == hash(
                        d_r
                    ), "Hash equality is inconsistent with __eq__"


def test_hashing_of_device():
    """
    Test that a :class:`dpctl.SyclDevice` object can be used as
    a dictionary key.

    """
    device_dict = {dpctl.SyclDevice(): "default_device"}
    assert device_dict


def test_equal():
    d1 = dpctl.SyclDevice()
    d2 = dpctl.SyclDevice()
    assert d1 != Ellipsis
    assert d1 == d2


list_of_supported_aspects = [
    "cpu",
    "gpu",
    "accelerator",
    "custom",
    "fp16",
    "fp64",
    "image",
    "online_compiler",
    "online_linker",
    "queue_profiling",
    "usm_device_allocations",
    "usm_host_allocations",
    "usm_shared_allocations",
    "usm_system_allocations",
    "host_debuggable",
    "atomic64",
    "usm_atomic_host_allocations",
    "usm_atomic_shared_allocations",
]

# SYCL 2020 spec aspects not presently
# supported in DPC++, and dpctl
list_of_unsupported_aspects = [
    "emulated",
]


@pytest.fixture(params=list_of_supported_aspects)
def supported_aspect(request):
    return request.param


@pytest.fixture(params=list_of_unsupported_aspects)
def unsupported_aspect(request):
    return request.param


def test_supported_aspect(supported_aspect):
    try:
        d = dpctl.SyclDevice()
        has_it = getattr(d, "has_aspect_" + supported_aspect)
    except dpctl.SyclDevieCreationError:
        has_it = False
    try:
        d_wa = dpctl.select_device_with_aspects(supported_aspect)
        assert getattr(d_wa, "has_aspect_" + supported_aspect)
    except dpctl.SyclDeviceCreationError:
        # ValueError may be raised if no device with
        # requested aspect charateristics is available
        assert not has_it


def test_unsupported_aspect(unsupported_aspect):
    try:
        dpctl.select_device_with_aspects(unsupported_aspect)
        raise AttributeError(
            f"The {unsupported_aspect} aspect is now supported in dpctl"
        )
    except AttributeError:
        pytest.skip(
            f"The {unsupported_aspect} aspect is not supported in dpctl"
        )


def test_handle_no_device():
    with pytest.raises(dpctl.SyclDeviceCreationError):
        dpctl.select_device_with_aspects(["gpu", "cpu"])
    with pytest.raises(dpctl.SyclDeviceCreationError):
        dpctl.select_device_with_aspects("cpu", excluded_aspects="cpu")


def test_cpython_api_SyclDevice_GetDeviceRef():
    import ctypes
    import sys

    d = dpctl.SyclDevice()
    mod = sys.modules[d.__class__.__module__]
    # get capsule storing SyclDevice_GetDeviceRef function ptr
    d_ref_fn_cap = mod.__pyx_capi__["SyclDevice_GetDeviceRef"]
    # construct Python callable to invoke "SyclDevice_GetDeviceRef"
    cap_ptr_fn = ctypes.pythonapi.PyCapsule_GetPointer
    cap_ptr_fn.restype = ctypes.c_void_p
    cap_ptr_fn.argtypes = [ctypes.py_object, ctypes.c_char_p]
    d_ref_fn_ptr = cap_ptr_fn(
        d_ref_fn_cap, b"DPCTLSyclDeviceRef (struct PySyclDeviceObject *)"
    )
    callable_maker = ctypes.PYFUNCTYPE(ctypes.c_void_p, ctypes.py_object)
    get_device_ref_fn = callable_maker(d_ref_fn_ptr)

    r2 = d.addressof_ref()
    r1 = get_device_ref_fn(d)
    assert r1 == r2


def test_cpython_api_SyclDevice_Make():
    import ctypes
    import sys

    d = dpctl.SyclDevice()
    mod = sys.modules[d.__class__.__module__]
    # get capsule storign SyclContext_Make function ptr
    make_d_fn_cap = mod.__pyx_capi__["SyclDevice_Make"]
    # construct Python callable to invoke "SyclDevice_Make"
    cap_ptr_fn = ctypes.pythonapi.PyCapsule_GetPointer
    cap_ptr_fn.restype = ctypes.c_void_p
    cap_ptr_fn.argtypes = [ctypes.py_object, ctypes.c_char_p]
    make_d_fn_ptr = cap_ptr_fn(
        make_d_fn_cap, b"struct PySyclDeviceObject *(DPCTLSyclDeviceRef)"
    )
    callable_maker = ctypes.PYFUNCTYPE(ctypes.py_object, ctypes.c_void_p)
    make_d_fn = callable_maker(make_d_fn_ptr)

    d2 = make_d_fn(d.addressof_ref())
    assert d == d2



#                      Data Parallel Control (dpctl)
#
# Copyright 2020-2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import contextmanager
from os import environ, getenv
from platform import system as sys_platform

__doc__ = (
    "Implementation module of :class:`dpctl.utils.onetrace_enabled` "
    "context manager."
)

_UNCHECKED = sys_platform() == "Linux"
del sys_platform


@contextmanager
def onetrace_enabled():
    """Enable `onetrace` collection for kernels executed in this context.

    N.B.: Proper working of this utility assumes that Python interpreter
    has been launched by `onetrace` tool from intel/pti-gpu project.

    :Example:
        Launch the Python interpreter using `onetrace` tool: ::

            $ onetrace --conditional-collection -v -t --demangle python app.py

        Now using the context manager in the Python sessions enables
        data collection and its output for every offloaded kernel ::

            import dpctl.tensor as dpt
            from dpctl.utils import onetrace_enabled

            # onetrace output reporting on execution of the kernel
            # should be seen, starting with "Device Timeline"
            with onetrace_enabled():
                dpt.arange(100, dtype='int16')

    """
    global _UNCHECKED

    if _UNCHECKED:
        _UNCHECKED = False
        if not (
            getenv("PTI_ENABLE", None) == "1"
            and "onetrace_tool" in getenv("LD_PRELOAD", "")
        ):
            import warnings

            warnings.warn(
                "It looks like Python interpreter was not started using "
                "`onetrace` utility. Using `onetrace_enabled` may have "
                "no effect. See `onetrace_enabled.__doc__` for usage.",
                RuntimeWarning,
                stacklevel=2,
            )

    _env_var_name = "PTI_ENABLE_COLLECTION"
    saved = getenv(_env_var_name, None)
    try:
        environ[_env_var_name] = "1"
        yield
    finally:
        if saved is None:
            del environ[_env_var_name]
        else:
            environ[_env_var_name] = saved



#                      Data Parallel Control (dpctl)
#
# Copyright 2020-2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import dpctl

list_of_standard_selectors = [
    dpctl.select_accelerator_device,
    dpctl.select_cpu_device,
    dpctl.select_default_device,
    dpctl.select_gpu_device,
]

list_of_valid_filter_selectors = [
    "opencl",
    "opencl:gpu",
    "opencl:cpu",
    "opencl:gpu:0",
    "gpu",
    "cpu",
    "level_zero",
    "level_zero:gpu",
    "opencl:cpu:0",
    "level_zero:gpu:0",
    "gpu:0",
    "gpu:1",
    "0",
]

list_of_invalid_filter_selectors = [
    "-1",
    "opencl:gpu:-1",
    "cuda:cpu:0",
    "abc",
    1,
]


# Unit test cases that will be run for every device
def check_max_compute_units(device):
    max_compute_units = device.max_compute_units
    assert max_compute_units > 0


def check_global_mem_size(device):
    global_mem_size = device.global_mem_size
    assert global_mem_size > 0


def check_local_mem_size(device):
    local_mem_size = device.local_mem_size
    assert local_mem_size > 0


def check_max_work_item_dims(device):
    max_work_item_dims = device.max_work_item_dims
    assert max_work_item_dims > 0


def check_max_work_item_sizes1d(device):
    max_work_item_sizes = device.max_work_item_sizes1d
    for size in max_work_item_sizes:
        assert size is not None


def check_max_work_item_sizes2d(device):
    max_work_item_sizes = device.max_work_item_sizes2d
    for size in max_work_item_sizes:
        assert size is not None


def check_max_work_item_sizes3d(device):
    max_work_item_sizes = device.max_work_item_sizes3d
    for size in max_work_item_sizes:
        assert size is not None


def check_max_work_item_sizes(device):
    with pytest.warns(DeprecationWarning):
        max_work_item_sizes = device.max_work_item_sizes
    for size in max_work_item_sizes:
        assert size is not None


def check_max_work_group_size(device):
    max_work_group_size = device.max_work_group_size
    # Special case for FPGA simulator
    if device.is_accelerator:
        assert max_work_group_size >= 0
    else:
        assert max_work_group_size > 0


def check_max_num_sub_groups(device):
    max_num_sub_groups = device.max_num_sub_groups
    # Special case for FPGA simulator
    if device.is_accelerator:
        assert max_num_sub_groups >= 0
    else:
        assert max_num_sub_groups > 0


def check_sub_group_sizes(device):
    sg_sizes = device.sub_group_sizes
    assert all(el > 0 for el in sg_sizes)


def check_has_aspect_cpu(device):
    try:
        device.has_aspect_cpu
    except Exception:
        pytest.fail("has_aspect_cpu call failed")


def check_has_aspect_gpu(device):
    try:
        device.has_aspect_gpu
    except Exception:
        pytest.fail("has_aspect_gpu call failed")


def check_has_aspect_accelerator(device):
    try:
        device.has_aspect_accelerator
    except Exception:
        pytest.fail("has_aspect_accelerator call failed")


def check_has_aspect_custom(device):
    try:
        device.has_aspect_custom
    except Exception:
        pytest.fail("has_aspect_custom call failed")


def check_has_aspect_fp16(device):
    try:
        device.has_aspect_fp16
    except Exception:
        pytest.fail("has_aspect_fp16 call failed")


def check_has_aspect_fp64(device):
    try:
        device.has_aspect_fp64
    except Exception:
        pytest.fail("has_aspect_fp64 call failed")


def check_has_aspect_atomic64(device):
    try:
        device.has_aspect_atomic64
    except Exception:
        pytest.fail("has_aspect_atomic64 call failed")


def check_has_aspect_image(device):
    try:
        device.has_aspect_image
    except Exception:
        pytest.fail("has_aspect_image call failed")


def check_has_aspect_online_compiler(device):
    try:
        device.has_aspect_online_compiler
    except Exception:
        pytest.fail("has_aspect_online_compiler call failed")


def check_has_aspect_online_linker(device):
    try:
        device.has_aspect_online_linker
    except Exception:
        pytest.fail("has_aspect_online_linker call failed")


def check_has_aspect_queue_profiling(device):
    try:
        device.has_aspect_queue_profiling
    except Exception:
        pytest.fail("has_aspect_queue_profiling call failed")


def check_has_aspect_usm_device_allocations(device):
    try:
        device.has_aspect_usm_device_allocations
    except Exception:
        pytest.fail("has_aspect_usm_device_allocations call failed")


def check_has_aspect_usm_host_allocations(device):
    try:
        device.has_aspect_usm_host_allocations
    except Exception:
        pytest.fail("has_aspect_usm_host_allocations call failed")


def check_has_aspect_usm_shared_allocations(device):
    try:
        device.has_aspect_usm_shared_allocations
    except Exception:
        pytest.fail("has_aspect_usm_shared_allocations call failed")


def check_has_aspect_usm_restricted_shared_allocations(device):
    try:
        device.has_aspect_usm_restricted_shared_allocations
    except Exception:
        pytest.fail("has_aspect_usm_restricted_shared_allocations call failed")


def check_has_aspect_usm_system_allocations(device):
    try:
        device.has_aspect_usm_system_allocations
    except Exception:
        pytest.fail("has_aspect_usm_system_allocations call failed")


def check_has_aspect_usm_atomic_host_allocations(device):
    try:
        device.has_aspect_usm_atomic_host_allocations
    except Exception:
        pytest.fail("has_aspect_usm_atomic_host_allocations call failed")


def check_has_aspect_usm_atomic_shared_allocations(device):
    try:
        device.has_aspect_usm_atomic_shared_allocations
    except Exception:
        pytest.fail("has_aspect_usm_atomic_shared_allocations call failed")


def check_has_aspect_host_debuggable(device):
    try:
        device.has_aspect_host_debuggable
    except Exception:
        pytest.fail("has_aspect_host_debuggable call failed")


def check_is_accelerator(device):
    try:
        device.is_accelerator
    except Exception:
        pytest.fail("is_accelerator call failed")


def check_is_cpu(device):
    try:
        device.is_cpu
    except Exception:
        pytest.fail("is_cpu call failed")


def check_is_gpu(device):
    try:
        device.is_gpu
    except Exception:
        pytest.fail("is_gpu call failed")


def check_max_read_image_args(device):
    try:
        device.max_read_image_args
    except Exception:
        pytest.fail("max_read_image_args call failed")


def check_max_write_image_args(device):
    try:
        device.max_write_image_args
    except Exception:
        pytest.fail("max_write_image_args call failed")


def check_image_2d_max_width(device):
    try:
        device.image_2d_max_width
    except Exception:
        pytest.fail("image_2d_max_width call failed")


def check_image_2d_max_height(device):
    try:
        device.image_2d_max_height
    except Exception:
        pytest.fail("image_2d_max_height call failed")


def check_image_3d_max_width(device):
    try:
        device.image_3d_max_width
    except Exception:
        pytest.fail("image_3d_max_width call failed")


def check_image_3d_max_height(device):
    try:
        device.image_3d_max_height
    except Exception:
        pytest.fail("image_3d_max_height call failed")


def check_image_3d_max_depth(device):
    try:
        device.image_3d_max_depth
    except Exception:
        pytest.fail("image_3d_max_depth call failed")


def check_sub_group_independent_forward_progress(device):
    try:
        device.sub_group_independent_forward_progress
    except Exception:
        pytest.fail("sub_group_independent_forward_progress call failed")


def check_preferred_vector_width_char(device):
    try:
        device.preferred_vector_width_char
    except Exception:
        pytest.fail("preferred_vector_width_char call failed")


def check_preferred_vector_width_short(device):
    try:
        device.preferred_vector_width_short
    except Exception:
        pytest.fail("preferred_vector_width_short call failed")


def check_preferred_vector_width_int(device):
    try:
        device.preferred_vector_width_int
    except Exception:
        pytest.fail("preferred_vector_width_int call failed")


def check_preferred_vector_width_long(device):
    try:
        device.preferred_vector_width_long
    except Exception:
        pytest.fail("preferred_vector_width_long call failed")


def check_preferred_vector_width_float(device):
    try:
        device.preferred_vector_width_float
    except Exception:
        pytest.fail("preferred_vector_width_float call failed")


def check_preferred_vector_width_double(device):
    try:
        device.preferred_vector_width_double
    except Exception:
        pytest.fail("preferred_vector_width_double call failed")


def check_preferred_vector_width_half(device):
    try:
        device.preferred_vector_width_half
    except Exception:
        pytest.fail("preferred_vector_width_half call failed")


def check_native_vector_width_char(device):
    try:
        device.native_vector_width_char
    except Exception:
        pytest.fail("native_vector_width_char call failed")


def check_native_vector_width_short(device):
    try:
        device.native_vector_width_short
    except Exception:
        pytest.fail("native_vector_width_short call failed")


def check_native_vector_width_int(device):
    try:
        device.native_vector_width_int
    except Exception:
        pytest.fail("native_vector_width_int call failed")


def check_native_vector_width_long(device):
    try:
        device.native_vector_width_long
    except Exception:
        pytest.fail("native_vector_width_long call failed")


def check_native_vector_width_float(device):
    try:
        device.native_vector_width_float
    except Exception:
        pytest.fail("native_vector_width_float call failed")


def check_native_vector_width_double(device):
    try:
        device.native_vector_width_double
    except Exception:
        pytest.fail("native_vector_width_double call failed")


def check_native_vector_width_half(device):
    try:
        device.native_vector_width_half
    except Exception:
        pytest.fail("native_vector_width_half call failed")


def check_create_sub_devices_equally(device):
    try:
        n = int(device.max_compute_units / 2)
        device.create_sub_devices(partition=n)
    except dpctl.SyclSubDeviceCreationError:
        pytest.skip(
            "create_sub_devices can't create sub-devices on this device"
        )
    except Exception:
        pytest.fail("create_sub_devices failed")


def check_create_sub_devices_equally_zeros(device):
    try:
        device.create_sub_devices(partition=0)
    except TypeError:
        pass


def check_create_sub_devices_by_counts(device):
    try:
        n = device.max_compute_units / 2
        device.create_sub_devices(partition=(n, n))
    except dpctl.SyclSubDeviceCreationError:
        pytest.skip(
            "create_sub_devices can't create sub-devices on this device"
        )
    except Exception:
        pytest.fail("create_sub_devices failed")


def check_create_sub_devices_by_counts_zeros(device):
    try:
        device.create_sub_devices(partition=(0, 1))
    except TypeError:
        pass


def check_create_sub_devices_by_affinity_not_applicable(device):
    try:
        device.create_sub_devices(partition="not_applicable")
    except dpctl.SyclSubDeviceCreationError:
        pytest.skip(
            "create_sub_devices can't create sub-devices on this device"
        )
    except Exception:
        pytest.fail("create_sub_devices failed")


def check_create_sub_devices_by_affinity_numa(device):
    try:
        device.create_sub_devices(partition="numa")
    except dpctl.SyclSubDeviceCreationError:
        pytest.skip(
            "create_sub_devices can't create sub-devices on this device"
        )
    except Exception:
        pytest.fail("create_sub_devices failed")


def check_create_sub_devices_by_affinity_L4_cache(device):
    try:
        device.create_sub_devices(partition="L4_cache")
    except dpctl.SyclSubDeviceCreationError:
        pytest.skip(
            "create_sub_devices can't create sub-devices on this device"
        )
    except Exception:
        pytest.fail("create_sub_devices failed")


def check_create_sub_devices_by_affinity_L3_cache(device):
    try:
        device.create_sub_devices(partition="L3_cache")
    except dpctl.SyclSubDeviceCreationError:
        pytest.skip(
            "create_sub_devices can't create sub-devices on this device"
        )
    except Exception:
        pytest.fail("create_sub_devices failed")


def check_create_sub_devices_by_affinity_L2_cache(device):
    try:
        device.create_sub_devices(partition="L2_cache")
    except dpctl.SyclSubDeviceCreationError:
        pytest.skip(
            "create_sub_devices can't create sub-devices on this device"
        )
    except Exception:
        pytest.fail("create_sub_devices failed")


def check_create_sub_devices_by_affinity_L1_cache(device):
    try:
        device.create_sub_devices(partition="L1_cache")
    except dpctl.SyclSubDeviceCreationError:
        pytest.skip(
            "create_sub_devices can't create sub-devices on this device"
        )
    except Exception:
        pytest.fail("create_sub_devices failed")


def check_create_sub_devices_by_affinity_next_partitionable(device):
    try:
        device.create_sub_devices(partition="next_partitionable")
    except dpctl.SyclSubDeviceCreationError:
        pytest.skip(
            "create_sub_devices can't create sub-devices on this device"
        )
    except Exception:
        pytest.fail("create_sub_devices failed")


def check_print_device_info(device):
    try:
        device.print_device_info()
    except Exception:
        pytest.fail("Encountered an exception inside print_device_info().")


def check_repr(device):
    assert type(repr(device)) is str


def check_profiling_timer_resolution(device):
    try:
        resol = device.profiling_timer_resolution
    except Exception:
        pytest.fail(
            "Encountered an exception inside "
            "profiling_timer_resolution property."
        )
    assert isinstance(resol, int) and resol > 0


def check_platform(device):
    p = device.sycl_platform
    assert isinstance(p, dpctl.SyclPlatform)


def check_parent_device(device):
    pd = device.parent_device
    assert pd is None or isinstance(pd, dpctl.SyclDevice)


def check_partition_max_sub_devices(device):
    max_part = device.partition_max_sub_devices
    assert isinstance(max_part, int)
    assert max_part >= 0
    assert max_part <= device.max_compute_units


def check_filter_string(device):
    try:
        fs = device.filter_string
        assert type(fs) is str
        dd = dpctl.SyclDevice(fs)
        assert device == dd
    except TypeError:
        pass


def check_name(device):
    dn = device.name
    assert dn
    assert type(dn) is str


def check_driver_version(device):
    dv = device.driver_version
    assert dv
    assert type(dv) is str


def check_vendor(device):
    ve = device.vendor
    assert ve
    assert type(ve) is str


def check_default_selector_score(device):
    sc = device.default_selector_score
    assert type(sc) is int
    assert sc > 0 or not (
        device.has_aspect_cpu
        or device.has_aspect_gpu
        or device.has_aspect_accelerator
    )


def check_backend(device):
    be = device.backend
    assert type(be) is dpctl.backend_type


def check_device_type(device):
    dt = device.device_type
    assert type(dt) is dpctl.device_type


def check_global_mem_cache_type(device):
    gmc_ty = device.global_mem_cache_type
    assert type(gmc_ty) is dpctl.global_mem_cache_type


def check_global_mem_cache_size(device):
    gmc_sz = device.global_mem_cache_size
    assert type(gmc_sz) is int
    assert gmc_sz


def check_global_mem_cache_line_size(device):
    gmc_sz = device.global_mem_cache_line_size
    assert type(gmc_sz) is int
    assert gmc_sz


list_of_checks = [
    check_max_compute_units,
    check_max_work_item_dims,
    check_max_work_item_sizes1d,
    check_max_work_item_sizes2d,
    check_max_work_item_sizes3d,
    check_max_work_item_sizes,
    check_max_work_group_size,
    check_max_num_sub_groups,
    check_sub_group_sizes,
    check_is_accelerator,
    check_is_cpu,
    check_is_gpu,
    check_sub_group_independent_forward_progress,
    check_preferred_vector_width_char,
    check_preferred_vector_width_short,
    check_preferred_vector_width_int,
    check_preferred_vector_width_long,
    check_preferred_vector_width_float,
    check_preferred_vector_width_double,
    check_preferred_vector_width_half,
    check_native_vector_width_char,
    check_native_vector_width_short,
    check_native_vector_width_int,
    check_native_vector_width_long,
    check_native_vector_width_float,
    check_native_vector_width_double,
    check_native_vector_width_half,
    check_has_aspect_cpu,
    check_has_aspect_gpu,
    check_has_aspect_accelerator,
    check_has_aspect_custom,
    check_has_aspect_fp16,
    check_has_aspect_fp64,
    check_has_aspect_atomic64,
    check_has_aspect_image,
    check_has_aspect_online_compiler,
    check_has_aspect_online_linker,
    check_has_aspect_queue_profiling,
    check_has_aspect_usm_device_allocations,
    check_has_aspect_usm_host_allocations,
    check_has_aspect_usm_shared_allocations,
    check_has_aspect_usm_restricted_shared_allocations,
    check_has_aspect_usm_system_allocations,
    check_has_aspect_usm_atomic_host_allocations,
    check_has_aspect_usm_atomic_shared_allocations,
    check_has_aspect_host_debuggable,
    check_max_read_image_args,
    check_max_write_image_args,
    check_image_2d_max_width,
    check_image_2d_max_height,
    check_image_3d_max_width,
    check_image_3d_max_height,
    check_image_3d_max_depth,
    check_create_sub_devices_equally,
    check_create_sub_devices_by_counts,
    check_create_sub_devices_by_affinity_not_applicable,
    check_create_sub_devices_by_affinity_numa,
    check_create_sub_devices_by_affinity_L4_cache,
    check_create_sub_devices_by_affinity_L3_cache,
    check_create_sub_devices_by_affinity_L2_cache,
    check_create_sub_devices_by_affinity_L1_cache,
    check_create_sub_devices_by_affinity_next_partitionable,
    check_print_device_info,
    check_repr,
    check_global_mem_size,
    check_local_mem_size,
    check_profiling_timer_resolution,
    check_platform,
    check_parent_device,
    check_partition_max_sub_devices,
    check_filter_string,
    check_vendor,
    check_driver_version,
    check_name,
    check_default_selector_score,
    check_backend,
    check_device_type,
    check_global_mem_cache_type,
    check_global_mem_cache_size,
    check_global_mem_cache_line_size,
]


@pytest.fixture(params=list_of_valid_filter_selectors)
def valid_filter(request):
    return request.param


@pytest.fixture(params=list_of_invalid_filter_selectors)
def invalid_filter(request):
    return request.param


@pytest.fixture(params=list_of_standard_selectors)
def device_selector(request):
    return request.param


@pytest.fixture(params=list_of_checks)
def check(request):
    return request.param



#                      Data Parallel Control (dpctl)
#
# Copyright 2020-2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Configures pytest to discover helper/ module
"""

import os
import sys

from _device_attributes_checks import (
    check,
    device_selector,
    invalid_filter,
    valid_filter,
)

sys.path.append(os.path.join(os.path.dirname(__file__), "helper"))

# common fixtures
__all__ = ["check", "device_selector", "invalid_filter", "valid_filter"]



#                       Data Parallel Control (dpctl)
#
#  Copyright 2020-2023 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from numpy import dtype

bool = dtype("bool")
int8 = dtype("int8")
int16 = dtype("int16")
int32 = dtype("int32")
int64 = dtype("int64")
uint8 = dtype("uint8")
uint16 = dtype("uint16")
uint32 = dtype("uint32")
uint64 = dtype("uint64")
float16 = dtype("float16")
float32 = dtype("float32")
float64 = dtype("float64")
complex64 = dtype("complex64")
complex128 = dtype("complex128")

__all__ = [
    "dtype",
    "bool",
    "int8",
    "uint8",
    "int16",
    "uint16",
    "int32",
    "uint32",
    "int64",
    "uint64",
    "float16",
    "float32",
    "float64",
    "complex64",
    "complex128",
]
